package HSPC::MT::Plugin::PP::OP_WireCard;

use strict;
use HSPC::PluginToolkit::General qw(string log log_debug log_warn throw_exception 
translit_data_utf2ascii);

use LWP::UserAgent;
use MIME::Base64;
use Data::Dumper;

use HSPC::MT::Plugin::PP::OP_WireCard::Request;
use HSPC::MT::Plugin::PP::OP_WireCard::Response;


## Fields to be shadow while logging
use constant FIELD2SHADOW => qw(CreditCardNumber CVC2 BusinessCaseSignature);

## timeout for gateway response in seconds
use constant	CONNECT_TIMEOUT		=> 15;
## timeout for alarm (to handle "frozen" connection, in seconds)
use constant	ALARM_TIMEOUT		=> 30;


sub get_title {
	my $class = shift;
	
	return	string('wirecard_naming');
}

sub get_supported_payment_method_types{
	my $self = shift;
	my %h	= (@_);

	return [
		'M',	## mastercard
		'V',	## visa
		'A',	## american express
		'C',	## dinner club
		'E',	## enroute
		'B',	## dinersclub carteblanche
		'D',	## discover
		'J',	## jcb
		'W',	## switch
		'S',	## solo
		'T',	## delta
		'O',	## optima
		'L',	## electron
	];
}

sub _check_avs {
	my $self = shift;
	my %h = @_;

	my $reply	= $h{reply};
	my $config	= $h{config};

	return 0 if $config->{is_avs_enabled} && $reply->avs =~ /[AZNWU]/;

	return 1;
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_preauthorize {
	my $self = shift;
	my %h = @_;

	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $total_paid		= $h{transaction_amount};

	throw_exception (
		type		=> 'error',
		log_message	=> {
			string_id	=> 'wirecard_total_error',
			string_args	=> {total_paid => $total_paid},
		},
	)
		unless ( $total_paid && ($total_paid > 0) );
	
	my $error	= undef;
	my $result	= {};
	my $details	= {trans_id=>$now_trans_id};

	## prepare preauth request
	my $preauth	= $self->prepare_request( %h, type => 'PREAUTHORIZATION' );

	my $reply	= $self->submit_request( config => $config, request => $preauth );
	## handle "bad reply" (it throws exception)
	$self->bad_reply( request => $preauth, reply => $reply )	unless ref($reply);

	## set transaction details
	$self->set_details (
		config	=> $config,
		details	=> $details,
		reply	=> $reply,
	);

	if( $reply->result eq 'ACK' ) {
		## preauth OK
		if (
			!$self->_check_avs(
				reply  => $reply,
				config => $config,
			)
		) {
			## AVS verification failed for preauth
			log_warn "[CCP] WireCard: AVS verification failed: " . $reply->avs;

			## process reversal transaction
			$h{transaction_id}				= $now_trans_id;
			$h{previous_transaction_data}	= $details;
			$self->_process_reversal( %h );

			my $msg	= string( 'wirecard_avs_failed', avs => $reply->avs );
			$result	= {
				STATUS			=> 'FRAUD',
				TEXT			=> {
					customer_message	=> $msg,
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	=> {
					ccp_error	=> $msg,
					%$details,
				},
			};
		} else {
			my $msg	= string('wirecard_trans_approved');
			## SUCCESS
			$result = {
				STATUS	=> 'APPROVED',
				TEXT			=> {
					customer_message	=> $msg,
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	=> {
					amount		=> $total_paid,
					%$details,
				},
			};
		}
	} elsif( $reply->result eq 'PENDING' ) {
		## PENDING
		$details->{wirecard_check_status_count} = 0;
		my $msg	= string('wirecard_transaction_pending');
		$result = {
			STATUS	=> 'PENDING',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
			NEXT_TRANSACTION_GAP => $config->{check_status_period},
		};
	} elsif( $reply->result eq 'NOK' && $reply->err_type eq "REJECTED" ) {
		$result	= {
			STATUS              => 'DECLINED',
			TRANSACTION_DETAILS => $details
		};
	} elsif( $reply->err_message() ){
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		log_warn "[CCP] WireCard: Preauth transaction failed: " . $error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	}

	### check result
	if( defined $result->{STATUS} ) {
		return	$result;
	} else {
		my $msg	= string('unexpected_error_trying_to_process_payment'),
		throw_exception (
			type		=> 'error',
			log_message	=> {
				string_id	=> 'unexpected_error_trying_to_process_payment',
			},
		);
	}
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_preauthorize_void {
	my $self = shift;

	return $self->_process_reversal(@_);
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_capture {
	my $self	= shift;
	my %h		= @_;

	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $total_paid		= $h{transaction_amount};
	my $previous_transaction_data	= $h{previous_transaction_data};

	throw_exception (
		type		=> 'error',
		log_message	=> {
			string_id	=> 'wirecard_total_error',
			string_args	=> {total_paid => $total_paid},
		},
	)
		unless ( $total_paid && ($total_paid > 0) );

	my $result = {};
	my $details = {
		trans_id => $now_trans_id,
		amount   => $total_paid,
	};

	## prepare capture request
	my $capture	= $self->prepare_request (
					%h,
					type		=> 'CAPTURING',
					guwid		=> $previous_transaction_data->{guwid},
					auth_code	=> $previous_transaction_data->{wirecard_auth_code},
	);
	my $reply	= $self->submit_request( config => $config, request => $capture );
	## handle "bad reply" (it throws exception)
	$self->bad_reply( request => $capture, reply => $reply )
		unless ref($reply);

	## set transaction details
	$self->set_details (
		config	=> $config,
		details	=> $details,
		reply	=> $reply,
	);

	if($reply->result eq 'ACK') {
		## capture OK set transaction details
		$self->set_details( config => $config, details => $details, reply => $reply );
		my $msg	= string('wirecard_trans_approved');
		$result	= {
			STATUS	=> 'APPROVED',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
		};
	} elsif( $reply->result eq 'NOK' && $reply->err_type eq "REJECTED" ) {
		$result	= {
			STATUS              => 'DECLINED',
			TRANSACTION_DETAILS => $details
		};
	} elsif( $reply->err_message() ) {
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		## capture FAILED
		log_warn "[CCP] WireCard: Capture transaction failed: " . $error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	} elsif($reply->result eq 'PENDING'){
		## PENDING
		$details->{wirecard_check_status_count} = 0;
		my $msg	= string('wirecard_transaction_pending');
		$result = {
			STATUS	=> 'PENDING',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
			NEXT_TRANSACTION_GAP => $config->{check_status_period},
		};
	}

	### check result
	if( defined $result->{STATUS} ) {
		return $result;
	} else {
		my $msg	= string('unexpected_error_trying_to_process_payment'),
		throw_exception (
			type		=> 'error',
			log_message	=> {
				string_id	=> 'unexpected_error_trying_to_process_payment',
			},
		);
	} 	
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_capture_void {
	my $self = shift;

	return $self->_process_reversal(@_);
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_sale {
	my $self = shift;
	my %h = @_;

	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $total_paid		= $h{transaction_amount};
	my $previous_transaction_data	= $h{previous_transaction_data};

	my $error	= undef;
	my $result	= {};
	my $details	= {trans_id=>$now_trans_id};

	## prepare preauth request
	my $sale	= $self->prepare_request( %h, type => 'TRANSACTION');
	my $reply	= $self->submit_request( config => $config, request => $sale );
	## handle "bad reply" (it throws exception)
	$self->bad_reply(request=>$sale,reply=>$reply) unless ref($reply);

	## set transaction details
	$self->set_details( config => $config, details => $details, reply => $reply );

	if($reply->result eq 'ACK') {
		## preauth OK
		if (
			!$self->_check_avs(
				reply  => $reply,
				config => $config,
			)
		) {
			## CVV verification failed for preauth
			log_warn "[CCP] WireCard: AVS verification failed: " . $reply->avs;
			my $msg	= string( 'wirecard_avs_failed', avs => $reply->avs );
			$result	= {
				STATUS	=> 'FRAUD',
				TEXT	=> {
					customer_message	=> string('incorrect_info'),
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	=> {
					ccp_error	=> $msg,
					%$details,
				},
			};
		} else {
			## SUCCESS
			my $msg	= string('wirecard_trans_approved');
			$result = {
				STATUS	=> 'APPROVED',
				TEXT	=> {
					customer_message	=> $msg,
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	=> {
					amount		=> $total_paid,
					%$details,
				},
			};
		}
	} elsif($reply->result eq 'PENDING') {
		$details->{wirecard_check_status_count} = 0;
		my $msg	= string('wirecard_transaction_pending');
		$result = {
			STATUS	=> 'PENDING',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	 =>{
					amount		=> $total_paid,
					%$details,
			},
			NEXT_TRANSACTION_GAP => $config->{check_status_period},
		};
	} elsif( $reply->result eq 'NOK' && $reply->err_type eq "REJECTED" ) {
		$result	= {
			STATUS              => 'DECLINED',
			TRANSACTION_DETAILS => $details
		};
	} elsif($reply->err_message) {
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		log_warn "[CCP] WireCard: Sale transaction failed: ".$error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	}

	### check result
	if( defined $result->{STATUS} ) {
		return $result;
	} else {
		throw_exception (
			type			=> 'error',
			log_message		=> string('unexpected_error_trying_to_process_payment'),
		);
	} 	
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub process_credit {
	my $self = shift;
	my %h = @_;

	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $total_paid		= $h{transaction_amount};
	my $old_details		= $h{previous_transaction_data};

	throw_exception (
		type		=> 'error',
		log_message	=> {
			string_id	=> 'wirecard_total_error',
			string_args	=> {total_paid => $total_paid},
		},
	)
		unless ( $total_paid && ($total_paid > 0) );

	my $result = {};
	my $details = { trans_id => $h{now_trans_id} };

	## prepare bookback request
	my $bookback	= $self->prepare_request (
		%h,
		type		=> 'BOOKBACK',
		guwid		=> $old_details->{guwid},
		auth_code	=> $old_details->{wirecard_auth_code},
	);

	## process request
	my $reply	= $self->submit_request( config => $config, request => $bookback );
	## handle "bad reply" (it throws exception)
	$self->bad_reply(request=>$bookback,reply=>$reply) unless ref($reply);

	## set transaction details
	$self->set_details( config => $config, details=>$details,reply=>$reply );

	if($reply->result eq 'ACK') {
		## SUCCESS
		my $msg	= string('wirecard_trans_approved');
		$result = {
			STATUS	=> 'APPROVED',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
		};
	} elsif($reply->err_message){
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		## ERROR
		log_warn "[CCP] WireCard: Reversal/Bookback transaction failed: ".$error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	} elsif($reply->result eq 'PENDING'){
		## PENDING
		$details->{wirecard_check_status_count} = 0;
		my $msg	= string('wirecard_transaction_pending');
		$result = {
			STATUS	=> 'PENDING',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	 =>{
					amount		=> $total_paid,
					%$details,
			},
			NEXT_TRANSACTION_GAP => $config->{check_status_period},
		};
	}

	### check result
	if (defined $result->{STATUS}) {
		return $result;
	} else {
		throw_exception (
			type		=> 'error',
			log_message	=> {
				string_id	=> 'unexpected_error_trying_to_process_payment',
			},
		);
	} 	
}

sub process_check_status{
	my $self = shift;
	my %h    = (
		config                    => undef,
		transaction_id            => undef,
		previous_transaction_data => undef,
		@_
	);
	my $config       = $h{config};
	my $now_trans_id = $h{transaction_id};
	my $old_details  = $h{previous_transaction_data};

	my $amount = $h{transaction_amount} > 0
		? $h{transaction_amount}
		: $old_details->{amount};

	my $result;

	my $notify	= $self->prepare_request (
		type						=> 'QUERY',
		transaction_amount			=> $amount,
		config						=> $h{config},
		transaction_id				=> $now_trans_id,
		account_info				=> $h{account_info},
		currency_iso				=> $h{currency_iso},
		payment_method	  			=> $h{payment_method},
		guwid						=> $old_details->{guwid},
		previous_transaction_data	=> $h{previous_transaction_data},
		auth_code					=> $old_details->{wirecard_auth_code},
	);

	## process request
	my $reply = $self->submit_request( config => $config, request => $notify );
	## handle "bad reply" (it throws exception)
	$self->bad_reply(request=>$notify,reply=>$reply) unless ref($reply);
	## set transaction details
	$self->set_details( config => $config, details=>$old_details, reply=>$reply );
	$old_details->{wirecard_check_status_count}++;

	if($reply->result eq 'ACK') {
		## SUCCESS
		my $msg	= string('wirecard_trans_approved');
		$result = {
			STATUS	=> 'APPROVED',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> $old_details,
		};
	} elsif( $reply->result eq 'NOK' && $reply->err_type eq "REJECTED" ) {
		## DECLINED
		$result	= {
			STATUS              => 'DECLINED',
			TRANSACTION_DETAILS => $old_details
		};
	} elsif($reply->result eq 'PENDING'){
		## PENDING
		if ($config->{check_status_period} > $old_details->{wirecard_check_status_count}) {
			my $msg	= string('wirecard_transaction_pending');
			$result = {
				STATUS	=> 'PENDING',
				TEXT	=> {
					customer_message	=> $msg,
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	 => $old_details,
				NEXT_TRANSACTION_GAP => $config->{check_status_period},
			};
		} else {
			## Waiting AUTH CALL
			log_warn "[CCP] WireCard: trans check is over";
	
			my $msg	= string('wirecard_authcall_trans_check_over');
			$result = {
				STATUS	=> 'AUTHCALL',
				TEXT	=> {
					customer_message	=> $msg,
					vendor_message		=> $msg,
				},
				TRANSACTION_DETAILS	=> $old_details,
			};
		}
	} elsif($reply->err_message){
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		## ERROR
		log_warn "[CCP] WireCard: Reversal/Bookback transaction failed: ".$error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	} 

	### check result
	if (defined $result->{STATUS}) {
		return $result;
	} else {
		throw_exception (
			type		=> 'error',
			log_message	=> {
				string_id	=> 'unexpected_error_trying_to_process_payment',
			},
		);
	} 	

}

## create and fill-in request object
## >> self
## =>> type		-- type of transaction
## =>> config   -- config
## ...  other stuff
## <    returns reference to result hash
sub prepare_request {
	my $self	= shift;
	my %h		= (@_);
	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $total_paid		= $h{transaction_amount};
	my $account_no		= $h{account_info}->{id};
	my $old_details		= $h{previous_transaction_data};
	my $document_info	= $h{document_info};

	# Initialize transaction structure
	my $request	= HSPC::MT::Plugin::PP::OP_WireCard::Request->new();

	## amount in smallest units
	my $amount_in_small = int( $total_paid * 100 );
	my $description = $document_info->{id} . ' ' .
							translit_data_utf2ascii(
							data => $document_info->{description}
							);

	$request->prepare (
		type			=> $h{type}, 
		bcs				=> $config->{business_cs},
		job_id			=> $now_trans_id,
		trans_id		=> $now_trans_id,
		guwid			=> $old_details->{guwid},
		auth_code		=> $old_details->{wirecard_auth_code},
		country_code	=> $config->{country_code},
		amount			=> $amount_in_small,
		currency		=> $h{currency_iso},
		is_avs			=> $config->{is_avs_enabled},
		is_cvv			=> $config->{is_cvv2_enabled},
		ccard			=> $ccard,
		billing_email	=> $h{account_info}->{billing_email},
		description		=> substr ($description, 0, 250),
	);

	log "[CCP] WireCard: $h{type} Request structure created";

	# Return the input structure
	return $request;
}

## submits request
## >> self
## =>> config	-- hash with configuration
## =>> request	-- reference to request structure
## <    reference to response data or string with error description
sub submit_request {
	my $self	= shift;
	my %h		= (@_);
	my $reply	= '';
	my $config	= $h{config};
	my $request	= $h{request};

	log "[CCP] WireCard: Submiting transaction to " . $config->{server_url};

	my $xml		= $request->xml();

	my $headers			= HTTP::Headers->new;
	my $authorization	= MIME::Base64::encode_base64( $config->{wc_name}.':'.$config->{passwd} );
	$headers			= new HTTP::Headers( Authorization => "BASIC $authorization" );

	my $ua	= new LWP::UserAgent( timeout => &CONNECT_TIMEOUT() );
	my $req	= new HTTP::Request 'POST', $config->{server_url}, $headers;

	$req->content_type( 'text/xml' );
	$req->content( $xml );
	
	log_debug "Request to gateway ($config->{server_url}):\n" . $self->_safe_dump($xml);

	my $res;
	eval {
		local $SIG{ALRM} = sub { die "alarm\n" };
		alarm &ALARM_TIMEOUT();
		$res	= $ua->request($req);
		alarm 0;
	};
	
	if (!$@) {
		$reply  = $res->content;
	}

	log_debug "submit_request: error in request to gateway:\n" . $res->error_as_HTML
		if $res->is_error;

	unless($reply) {
		log_warn "[CCP] WireCard: REPLY WAS NOT RECEIVED";
		return	string('wirecard_empty_reply');
	}

	log "[CCP] WireCard: REPLY ON ".$h{request}->type()." RECEIVED";
	log_debug "submit_request: reply:\n" . $reply;

	## response object
	my $response	= HSPC::MT::Plugin::PP::OP_WireCard::Response->new();
	my $parse_ok	= $response->parse( xml => $reply );

	if( !$parse_ok || !$response->{type} ) {
		log_warn "[CCP] WireCard: Unrecognized reply was received: $reply";
		return	string('wirecard_cant_parse_reply');
	}

	# Return the response object
	return $response;
}

## forms result in case of no reply
## >> self
## =>> reply		-- reply from gateway
## =>> request		-- request to gateway
## <	result hash
sub bad_reply {
	my $self	= shift;
	my %h		= (@_);
	my $user_error	= string('wirecard_user_error');
	my $admin_error	= $h{reply} || $user_error;
	my $details		= {
		trans_id			=> $h{request}->trans_id,
		transaction_type	=> $h{request}->type,
	};
	throw_exception (
		type			=> 'error',
		log_message		=> $admin_error,
		vendor_message	=> $admin_error,
		user_message	=> $user_error,
	);
}

## set details
## >> self
## =>> config		-- configuration hash
## =>> details		-- details hash (optional, will be refilled if any will be passed)
## =>> reply		-- reply from gateway
## <	details hash
sub set_details {
	my $self	= shift;
	my %h		= (@_);
	
	my $config	= $h{config};
	my $details	= $h{details}	|| {};
	my $reply	= $h{reply}		|| return $details;

	$details->{transaction_type}	= $reply->type;
	$details->{guwid}				= $reply->guwid;
	$details->{wirecard_auth_code}	= $reply->auth_code;
	$details->{ccp_avs_code}		= $reply->avs if $reply->avs;
	$details->{error}               = $reply->err_message if $reply->err_message;
	$details->{error_code}          = $reply->err_number  if $reply->err_number;

	return $details;
}

## support all currencies by default
sub get_currencies_supported{
	my $self = shift;
	my %h	= (@_);
	return [
				'CAD', # Canadian Dollar
				'USD', # US Dollar
				'GBP', # Pound Sterling
				'EUR', # Euro
				'JPY', # Yen
				'CHF', # Swiss Franc
		];
}

## >> self
## =>> config   -- configuration hash
## ...  other stuff
## <    returns reference to result hash
sub _process_reversal {
	my $self	= shift;
	my %h		= (@_);

	my $config			= $h{config};
	my $now_trans_id	= $h{transaction_id};
	my $ccard			= $h{payment_method};
	my $old_details		= $h{previous_transaction_data};
	my $total_paid		= $h{transaction_amount};

	throw_exception (
		type		=> 'error',
		log_message	=> {
			string_id	=> 'wirecard_total_error',
			string_args	=> {total_paid => $total_paid},
		},
	)
		unless ( $total_paid && ($total_paid > 0) );

	my $result = {};
	my $details = { trans_id => $h{now_trans_id} };

	## prepare reversal request
	my $reversal	= $self->prepare_request (
		%h,
		type		=> 'REVERSAL',
		guwid		=> $old_details->{guwid},
		auth_code	=> $old_details->{wirecard_auth_code},
	);

	## process requst
	my $reply	= $self->submit_request( config => $config, request => $reversal );
	## handle "bad reply" (it throws exception)
	$self->bad_reply(request=>$reversal,reply=>$reply) unless ref($reply);

	if($reply->result eq 'ACK') {
		## SUCCESS
		$self->set_details( config => $config, details => $details, reply => $reply );
		my $msg	= string('wirecard_trans_approved');
		$result	= {
			STATUS	=> 'APPROVED',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
		};
	} elsif( $reply->err_message() ){
		my $error_log	= $reply->error_log();	
		my $error_msg	= $reply->error_msg();	
		## ERROR
		log_warn "[CCP] WireCard: Reversal transaction failed: ".$error_log;
		throw_exception (
			type			=> 'error',
			log_message		=> $error_log,
			vendor_message	=> $error_msg,
			user_message	=> $error_msg,
		);
	} elsif($reply->result eq 'PENDING'){
		## PENDING
		$details->{wirecard_check_status_count} = 0;
		my $msg	= string('wirecard_transaction_pending');
		$result = {
			STATUS	=> 'PENDING',
			TEXT	=> {
				customer_message	=> $msg,
				vendor_message		=> $msg,
			},
			TRANSACTION_DETAILS	=> {
				amount		=> $total_paid,
				%$details,
			},
			NEXT_TRANSACTION_GAP => $config->{check_status_period},
		};
	}

	### check result
	if( defined $result->{STATUS} ) {
		return $result;
	} else {
		my $msg	= string('unexpected_error_trying_to_process_payment'),
		throw_exception (
			type		=> 'error',
			log_message	=> {
				string_id	=> 'unexpected_error_trying_to_process_payment',
			},
		);
	} 	
}

## The method cuts off confidential info from XML-message.
## >> self
## >> xml message
## <	xml message without confidential information.
sub _safe_dump {
	my $self	= shift;
	my $data	= shift;
	my $str		= $data;
	foreach my $field (&FIELD2SHADOW()) {
		$str	=~ s|<$field>\w+</$field>|<$field>xxxx</$field>|gs;
	}
	return $str;
}
1;
