package HSPC::MT::Plugin::PP::OP_WireCard::Request;

use strict;
use HSPC::PluginToolkit::General qw(log_warn);
use XML::Simple;


################################################################################
## constructor
## >> class
## <	object reference
sub new	{
	my $class = shift;
	my $self = bless({}, $class);
	return $self;
}

################################################################################
## return XML packet representation of request
## >> self
## <	string containing XML packet
sub xml {
	my $self	= shift;

	## header
	my $out	= '<?xml version="1.0" encoding="UTF-8" ?>' .
		'<WIRECARD_BXML xmlns:xsi="http://www.w3.org/1999/XMLSchema-instance"' .
		' xsi:noNamespaceSchemaLocation="wirecard.xsd">';

	## WireCard request
	$out	.= eval {
				XMLout (
					{
						W_REQUEST	=> $self->{W_REQUEST}
					},
					rootname	=> undef,
					noattr		=> 1,
					keeproot	=> 1,
				);
	};
	if( $@ ) {
		log_warn "[CCP] WireCard: ERROR while XMLout WireCard request: $@";
	}

	## footer
	$out	.= qq[</WIRECARD_BXML>];

	return $out;
}

################################################################################
## interface method for filling-in request object
## >> self
## =>> type				-- transaction type				[REQ]
## =>> bcs				-- business case signiture		[REQ]
## =>> job_id			-- JobID						[optional]
## =>> func_id			-- FunctionID					[optional]
## =>> trans_id			-- TransactionID				[REQ]
## =>> guwid			-- GuWID						[req for post-txn]
## =>> auth_code		-- AuthorizationCode			[req for post-txn]
## =>> country_code		-- country code					[REQ]
## =>> amount			-- amount of transaction		[REQ]
## =>> currency			-- currecny used				[REQ]
## =>> ccard			-- hash with credit card info	[req for booking txn]
## =>> billing_email	-- billing email
## <	undef
sub prepare {
	my $self = shift;
	my %h = (@_);

	$self->{type} = $h{type};
	$self->{trans_id} = $h{trans_id};

	## cc_transaction block
	my $cc_txn = {
		TransactionID	=> $h{trans_id},
		Amount			=> $h{amount},
		Currency		=> $h{currency},
		CountryCode		=> $h{country_code},
	};

	## optional parameters
	$cc_txn->{GuWID} = $h{guwid} if defined $h{guwid};
	$cc_txn->{AuthorizationCode} = $h{auth_code} if defined $h{auth_code};
	$cc_txn->{Usage} = $h{description} if defined $h{description};

	## information blocks for booking transactions ##
	if($h{type} eq 'TRANSACTION' || $h{type} eq 'PREAUTHORIZATION') {

		## credit card details
		if ($h{ccard}) {
			my $card_info = $self->_prepare_ccard(ccard=>$h{ccard},is_cvv=>$h{is_cvv});
			$cc_txn->{CREDIT_CARD_DATA} = [$card_info];
		}

		## contact info (currently only remote customer IP)
		if ($h{ip}) {
			my $ip_info = $self->_prepare_contact(ip=>$h{ip});
			$cc_txn->{CONTACT_DATA} = [$ip_info];
		}

		## AVS block
		if ($h{ccard}) {
			my $cc = $self->_prepare_addr (
				ccard			=> $h{ccard},
				billing_email	=> $h{billing_email},
			);
			$cc_txn->{CORPTRUSTCENTER_DATA} = [{ADDRESS => [$cc]}];
		}
	}

	## finally, arrange all structures into one WireCard request
	$self->{W_REQUEST}	= [
		{
			W_JOB	=> [
				{
					## if job_id was not defined, set it to trans_id or '1'
					JobID					=> $h{job_id} || $h{trans_id} || '1',
					BusinessCaseSignature	=> $h{bcs},
					"FNC_CC_$h{type}"		=> {
						FunctionID		=> ($h{func_id} || $h{type}),
						CC_TRANSACTION	=> [$cc_txn],
					},
				}
			],
		}
	];

	return	undef;
}

################################################################################
## private helper functions to fill-in request object

## >> self
## =>> ccard			-- hash with credit card data
## <	hash with credit card structure
sub _prepare_ccard {
	my ($self,%h)	= @_;
	my $ccard	= $h{ccard};
	my $sd		= $ccard->{secure_data};

	## expiration date
	my @exp = ($ccard->{expire_date} =~ m|^(\d+)\D+(\d+)|); ## ex. '2007-01-31 23:59:59'
	return {
		CreditCardNumber	=> $sd->{card_number},
		($h{is_cvv}) ? (CVC2 => $sd->{cvv}) : (),
		ExpirationYear		=> $exp[0],
		ExpirationMonth		=> sprintf("%02d", $exp[1]),
		CardHolderName		=> $sd->{holder_name}.' '.$sd->{holder_name2},
	};
}

## >> self
## =>> ip			-- customer ip
## <	hash with contact information
sub _prepare_contact {
	my ($self,%h) = @_;
	return { IPAddress => $h{ip} };
}

## prepares address structure
## >> self
## =>> ccard			-- hash with credit card data
## =>> billing_email	-- billing email
## <	hash with address structure
sub _prepare_addr {
	my ($self,%h) = @_;
	my $ccard	= $h{ccard};
	my $sd		= $ccard->{secure_data};

	my $addr	= {
		FirstName	=> $sd->{holder_name}, 
		LastName	=> $sd->{holder_name2},
		Address1	=> $sd->{address1},
		Address2	=> $sd->{address2},
		State		=> $sd->{state},
		City		=> $sd->{city},
		Country		=> $sd->{country},
		ZipCode		=> $sd->{zip},
		Phone		=> $sd->{phone},
		Email		=> $h{billing_email},
	};

	## replace empty values with "NoValue" string
	foreach (keys %$addr) {
		$addr->{$_} ||= 'No'.$_ unless defined $addr->{$_};
	}
	return $addr;
}

## PUBLIC ACCESSORS ##

sub type { return $_[0]->{type}; }
sub trans_id { return $_[0]->{trans_id}; }

1;

