package HSPC::MT::Plugin::PP::OP_WireCard::Response;

use strict;

use HSPC::PluginToolkit::General qw(log_warn string);
use XML::Simple;


###############################################################################
## response object constructor
## >> class
## =>> xml (optional)	- XML response packet to be parsed
## < object reference
sub new {
	my $class = shift;
	my %h = (@_);
	my $self = bless {}, $class;

	## parse response if passed
	$self->parse(xml=>$h{xml}) if $h{xml};
	
	return $self;
}

###############################################################################
## parse XML response packet
## >> self
## =>> xml #REQ
## < 1 on success, 0 - otherwise
sub parse {
	my $self = shift;
	my %h = (@_);
	return 0 unless $h{xml};

	## WireCard request
	my $response = eval { XMLin($h{xml},noattr=>1); };
	if($@) {
		log_warn "[CCP] WireCard: ERROR occurred while parsing XML response: $h{xml}: $@";
		return 0;
	}

	if($response && ref($response) eq 'HASH') {
		$self->{W_RESPONSE} = $response->{W_RESPONSE};
	}

	## detect transaction type
	my ($type) = grep { /FNC_CC/ } keys %{$self->{W_RESPONSE}->{W_JOB}};
	$type =~ s/^FNC_CC_//;
	$self->{type} = $type;

	return 1;
}

###############################################################################
sub error_msg {
	my $self = shift;
	if($self->err_type eq 'REJECTED') {
		return string('wirecard_trans_declined');
	} else {
		return string('unexpect_error_try_again');
	}
}
	
###############################################################################
sub error_log {
	my $self = shift;
	return	string (
		'wirecard_error_log',
		error_code	=> $self->err_number(),
		error_type	=> $self->err_type(),
		message		=> $self->err_message(),
		advice		=> $self->err_advice(),
	);
}

###############################################################################
## raw ERROR methods ##

## error message
sub err_message {
	my $self	= shift;
	my $errs	= $self->_error();

	my $msg;
	if( 'ARRAY' eq ref($errs) ) {
		my @msgs;
		foreach my $err (@$errs) {
			push( @msgs, $err->{Message} )	if $err->{Message};
		}
		$msg	= join( ' ', @msgs );
	} elsif( 'HASH' eq ref($errs) ) {
		$msg	= $errs->{Message};
	}
	return	$msg;
}

## error number
sub err_number {
	my $self	= shift;
	my $errs	= $self->_error();
	my $msg;
	if( 'ARRAY' eq ref($errs) ) {
		my @msgs;
		foreach my $err (@$errs) {
			push( @msgs, $err->{Number} )	if $err->{Number};
		}
		$msg	= join( ', ', @msgs );
	} elsif( 'HASH' eq ref($errs) ) {
		$msg	= $errs->{Number};
	}
	return	$msg;
}

## error type
sub err_type {
	my $self	= shift;
	my $errs	= $self->_error();
	my $msg;
	if( 'ARRAY' eq ref($errs) ) {
		my @msgs;
		foreach my $err (@$errs) {
			push( @msgs, $err->{Type} )	if $err->{Type};
		}	
		$msg	= join( ', ', @msgs );
	} elsif( 'HASH' eq ref($errs) ) {
		$msg	= $errs->{Type};
	}
	return	$msg;
}

## advices
sub err_advice {
	my $self	= shift;
	my $advs	= $self->_error();
	my $msg;
	if( 'ARRAY' eq ref($advs) ) {
		my @msgs;
		foreach my $adv (@$advs) {
			$adv	= $adv->{Advice};
			push( @msgs, ('ARRAY' eq ref($adv)) ? join("\n",@$adv) : $adv );
		}	
		$msg	= join( ' ', @msgs );
	} elsif( 'HASH' eq ref($advs) ) {
		my $adv	= $advs->{Advice};
		$msg	= ('ARRAY' eq ref($adv)) ? join("\n",@$adv) : $adv;
	}
	return	$msg;
}

###############################################################################
## accessor functions for convenient access

## return parsed XML (as is)
sub xml_response { return $_[0]->{W_RESPONSE}; }

## txn result
sub guwid     { return $_[0]->_status()->{GuWID}; }
sub result    { return $_[0]->_status()->{FunctionResult}; }
sub timestamp { return $_[0]->_status()->{TimeStamp}; }
sub info      { return $_[0]->_status()->{Info}; }

## auth
sub auth_code {
	my $str = $_[0]->_status()->{AuthorizationCode};
	return undef if ref $str;
	$str =~ s/\s//g;
	return $str;
}

## AVS code
sub avs {
	my $self = shift;
	my $avs = eval { 
		return $self->_cc_txn()->{CORPTRUSTCENTER_DATA}->{AVS}->{AVS_CODE};
	};
	return $@ ? undef : $avs;
}

sub type { my $self = shift; return $self->{type}; }

###############################################################################
## PRIVATE HELPER FUNCTIONS ##

## return processing status hashref
sub _cc_txn {
	my $self = shift;
	my $fnc = eval {
		return $self->{W_RESPONSE}->{W_JOB}->{'FNC_CC_'.$self->{type}}->{CC_TRANSACTION};
	};
	return $fnc || {};
}

## return processing status hashref
sub _status { return $_[0]->_cc_txn()->{PROCESSING_STATUS} || {}; }
sub _error  { return $_[0]->_status()->{ERROR} || {}; }

1;

